!---------------------------------------------------------------------------------------
! PREP_REGIONS: get_surfinf
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> get_surfinf
!!
!! Purpose:    Calculate total surface influence from flexpart grid_time files.
!!
!! Interface:
!!
!!    Inputs
!!             files    -  files data structure
!!             recs     -  receptor IDs for each observation
!!             obstimes -  observation time stamps in julian days
!!             surfinf  -  total surface influence (zero matrix on input)
!!
!!    Externals
!!             caldate
!!             read_grid
!!
!---------------------------------------------------------------------------------------

subroutine get_surfinf(files, config, recs, obstimes, surfinf)

  use mod_var
  use mod_dates
  use mod_settings
  use mod_flexpart

  implicit none

  type (files_t),                     intent (in)     :: files
  type (config_t),                    intent (in)     :: config
  real(kind=8), dimension(maxobs),    intent (in)     :: obstimes
  character(3), dimension(maxobs),    intent (in)     :: recs
  real, dimension(nxregrid,nyregrid), intent (in out) :: surfinf

  character(len=max_path_len)                         :: path_flexrec, filename, filedates
  character(len=6)                                    :: adate
  character(len=8)                                    :: areldate, areltime
  logical                                             :: lexist
  integer                                             :: ierr
  integer                                             :: jjjjmmdd, hhmiss
  real, dimension(nxgrid,nygrid,ngrid)                :: grid                       
  real(kind=8), dimension(ngrid)                      :: gtime             
  integer                                             :: i, ix1, ix2, jy1, jy2

  ! indices to regional domain
  ! --------------------------

  ix1 = minloc(glon, dim=1, mask=abs(glon-dx/2.-rllx).eq.minval(abs(glon-dx/2.-rllx)))
  ix2 = minloc(glon, dim=1, mask=abs(glon+dx/2.-rllx-nxregrid*rdx).eq.minval(abs(glon+dx/2.-rllx-nxregrid*rdx)))
  jy1 = minloc(glat, dim=1, mask=abs(glat-dy/2.-rlly).eq.minval(abs(glat-dy/2.-rlly)))
  jy2 = minloc(glat, dim=1, mask=abs(glat+dy/2.-rlly-nyregrid*rdy).eq.minval(abs(glat+dy/2.-rlly-nyregrid*rdy)))

  print*, 'get_surfinf: glon(ix1), glon(ix2), glat(jy1), glat(jy2) = ',glon(ix1), glon(ix2), glat(jy1), glat(jy2)

  ! loop over obs times
  ! -------------------

  do i = 1, nobs

    ! define file name
    call caldate(obstimes(i), jjjjmmdd, hhmiss)
    write(adate,'(I6.6)') jjjjmmdd/100
    write(areldate,'(I8.8)') jjjjmmdd
    write(areltime,'(I6.6)') hhmiss
    print*, 'jjjjmmdd = ',jjjjmmdd
    print*, 'hhmiss = ',hhmiss
    print*, 'recs(i) = ',trim(recs(i))
    print*, 'obstimes(i) = ',obstimes(i)
    path_flexrec = trim(files%path_flexpart)//trim(recs(i))//'/'//trim(adate)//'/'
    if (.not.config%nested)    filename = trim(path_flexrec)//'grid_time_'//trim(areldate)//trim(areltime)//'_001'
    if (config%nested) filename = trim(path_flexrec)//'grid_time_nest_'//trim(areldate)//trim(areltime)//'_001'
    filedates = trim(path_flexrec)//'dates'
    print*, 'get_surfinf: grid filename = ',filename

    ! read flexpart footprint
    inquire ( file=trim(filename),exist=lexist )
    if ( lexist ) then
      write(*,*) 'Reading flexpart :'//trim(filename)
      call read_grid(filename, filedates, obstimes(i), nxgrid, nygrid, grid, gtime)
      ! convert s.m3/kg to s.m2/kg
      grid = grid/outheight(1)
    else
      write(*,*) 'WARNING: cannot find '//trim(filename)
      go to 10
    endif

    surfinf(:,:) = surfinf(:,:) + sum(grid(ix1:ix2,jy1:jy2,:),dim=3)

10  continue
  
  end do

end subroutine get_surfinf

