!---------------------------------------------------------------------------------------
! FLEXINVERT: congrad
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> congrad
!! Lanczos algorithm for solving linear equations
!!
!! Pupose:     Minimizes the cost function by calculating the direction and magnitude 
!!             to update the state vector. Also calculates the eigenvalues and vectors
!!             of the Hessian.
!! Details:
!!
!! 
!! Interface:  
!!
!!    Inputs
!!             iter     -  number of iterations
!!             grad     -  gradient of the cost function
!!             files    -  file data structure
!!             fluxes   -  flux data structure
!!             obs      -  observation data structure
!!             states   -  state vector data structure
!!             covar    -  error covariance data structure 
!!
!!    External
!!             simulate
!!             dsyevr (LAPACK routine)            
!!
!! References: Lanczos, C., J. Res. Natl. Bur. Standards, 45, 255-282, 1950
!!             Paige, C. C. & Saunders, M. A., SIAM J. Numerical Aanalysis, 12(4), 1975
!!             Based on the code of Mike Fisher, ECMWF         
!
!---------------------------------------------------------------------------------------


subroutine congrad(iter, grad, files, config, fluxes, obs, states, covar)

  use mod_var
  use mod_settings
  use mod_fluxes
  use mod_obs
  use mod_states
  use mod_covar
  use mod_transform

  implicit none

  type (files_t),                  intent (in)     :: files
  type (config_t),                 intent (in)     :: config
  type (fluxes_t),                 intent (in)     :: fluxes
  type (obs_t),                    intent (in)     :: obs
  type (states_t),                 intent (in out) :: states
  type (covar_t),                  intent (in out) :: covar
  real(kind=8), dimension(nvar,1), intent (in out) :: grad
  integer,                         intent (in out) :: iter 

  real,         dimension(nvar)                    :: grad_o    ! observation gradient
  real                                             :: cost_o    ! cost observation space
  real(kind=8), dimension(nvar,1)                  :: x_ad      ! observation gradient physical space     
  real(kind=8), dimension(nvar,1)                  :: z_ad      ! observation gradient chi space
  real(kind=8), dimension(nvar,1)                  :: zgrad0    ! initial gradient
  real(kind=8), dimension(nvar,1)                  :: x0        ! prior state vector physical space
  real(kind=8), dimension(nvar,1)                  :: z0        ! prior state vector chi space
  real(kind=8), dimension(nvar,1)                  :: xbg       ! best guess at state vector physical space
  real(kind=8), dimension(nvar,1)                  :: zbg       ! best guess at state vector chi space
  real(kind=8), dimension(nvar,1)                  :: zw        ! lanczos vector
  real(kind=8), dimension(nvar,1)                  :: zlanc1    ! initial lanczos vector
  real(kind=8), dimension(nvar,maxiter+1)          :: zcglwk    ! matrix of lanczos vectors, V
  real(kind=8), dimension(nvar,1)                  :: xa        ! analysis state vector in physical space       
  real(kind=8), dimension(nvar,1)                  :: za        ! analysis state vector in chi space
  real(kind=8), dimension(nvar,1)                  :: zx        ! state vector in physical space
  real(kind=8), dimension(maxiter+1,1)             :: zqg0         
  real(kind=8), dimension(maxiter+1)               :: zbeta     ! vector of beta_j
  real(kind=8), dimension(maxiter)                 :: zdelta    ! vector of delta_j
  real(kind=8), dimension(maxiter+1,1)             :: zwork        
  real(kind=8), dimension(neig)                    :: tmp0
  real(kind=8), dimension(ndvar)                   :: tmp1
  real(kind=8), dimension(nvar)                    :: tmp
  real(kind=8)                                     :: dla          
  real(kind=8)                                     :: zgnorm    ! initial gradient norm
  real(kind=8)                                     :: reduct    ! reduction in gradient norm
  real(kind=8)                                     :: var       ! posterior variance
  real(kind=8), dimension(:,:), allocatable        :: zmat      ! tridiagonal matrix T
  real(kind=8), dimension(:,:), allocatable        :: izmat     ! inverse of matrix T
  real(kind=8), dimension(:,:), allocatable        :: zv        ! matrix of eigenvectors of T
  real(kind=8), dimension(:,:), allocatable        :: zritz     ! eigenvalues of T, Ritz values
  real(kind=8), dimension(:),   allocatable        :: ptheta    ! good eigenvalues of T, Ritz values
  real(kind=8), dimension(:,:), allocatable        :: pvcglev   ! matrix of good eigenvectors of T
  real(kind=8), dimension(:,:), allocatable        :: pevecs    ! matrix of eigenvectors of hessian
  integer,      dimension(:),   allocatable        :: isuppz
  real(kind=8), dimension(:),   allocatable        :: work
  integer,      dimension(:),   allocatable        :: iwork
  real(kind=8), dimension(:,:), allocatable        :: ieval
  real, dimension(nvar)                            :: delx
  real                                             :: cost_p, cost

  integer                                          :: jm, i, j, k, jj, ij, it, jk, jt 
  integer                                          :: ierr, ngood
  integer                                          :: lwork, liwork, info, m
  character(max_path_len)                          :: rowfmt
  character(len=2)                                 :: aiter

  ! initialize variables
  ! --------------------

  ! state vector physical space
  xbg(:,1) = dble(states%px) 

  ! prior state vector physical space
  x0(:,1) = dble(states%px0)

  ! transform prior state vector chi space
  z0 = phi2chi(states, covar, x0)

  ! transform state vector chi space
  zbg = phi2chi(states, covar, xbg)
 
  ! initial lanczos vector is initial gradient in chi space
  zgrad0(:,1) = grad(:,1)
  zlanc1(:,1) = grad(:,1)
  zgnorm = sqrt(dot_product(grad(:,1),grad(:,1)))

  ! initial lanczos vector
  zcglwk(:,1) = zlanc1(:,1)/sqrt(dot_product(zlanc1(:,1),zlanc1(:,1)))
  zqg0(1,1) = dot_product(zcglwk(:,1),zgrad0(:,1))

  ! Lanczos iteration
  ! -----------------

  lanczos_iter: do

    iter = iter + 1

    write(logid,*) 'Lanczos iteration: ',iter

    ! latest value of state vector in chi space
    zw(:,1) = zbg(:,1) + zcglwk(:,iter)

    ! calculate new gradient, A*v_j
    ! -----------------------------
    
    ! evaluate Hessian applied to latest value of zw

    if ( (config%restart.eq.1).and.(iter.le.lastiter) ) then
      ! use precalculated gradient
      write(aiter,fmt='(I2.2)') iter
      open(100,file=trim(files%path_output)//'grad_'//aiter//'.txt',action='read',status='old',iostat=ierr)
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR congrad: cannot open grad_'//aiter//'.txt'
        stop
      endif
      do i = 1, nvar
        read(100,*) grad_o(i)
      end do
      close(100)
    else
      ! transform zw to physical space
      zx = chi2phi(states, covar, zw)
      states%px = real(zx(:,1),kind=4)
      cost_o = 0.
      grad_o(:) = 0.
      call simulate(iter, files, config, fluxes, obs, states, grad_o, cost_o)
      ! calculate cost
      ! Jp = (p - p0)^TB^(-1)(p - p0) 
      delx = states%px - states%px0
      tmp = prod_icov(covar, delx)
      cost_p = dot_product( delx, real(tmp,kind=4) )
      ! J = 0.5*(Jp + Jo) 
      cost = 0.5*(cost_p + cost_o)
      write(aiter,'(I2.2)') iter
      write(logid,*) 'Cost iteration '//aiter//': ',cost
    endif

    ! transform gradient to chi space
    x_ad(:,1) = dble(grad_o)
    z_ad = phi2chi_ad(covar, x_ad )

    ! new gradient J' = J'p + J'o
    grad(:,1) = zw(:,1) + z_ad(:,1)

    ! subtract initial lanczos vector
    grad(:,1) = grad(:,1) - zgrad0(:,1)

    ! calculate new lanczos vector
    ! ----------------------------

    ! calculate zdelta a_j = < v_j+1, v_j >
    zdelta(iter) = dot_product( grad(:,1), zcglwk(:,iter) )
    write(logid,*) 'zdelta:',zdelta(iter)
    if ( zdelta(iter).le.0. ) then
      write(logid,*) 'ERROR: hessian is not positive definite, iter = ',iter
      stop
    endif

    ! calculate v_j+1 = v_j+1 - a_j*v_j
    grad(:,1) = grad(:,1) - zdelta(iter)*zcglwk(:,iter)

    ! calculate v_j+1 = A*v_j - B_j*v_j-1
    if( iter.gt.1 ) then
      grad(:,1) = grad(:,1) - zbeta(iter)*zcglwk(:,iter-1)
    endif

    ! orthonormalize gradient against other gradients
    do jm = iter,1,-1
      dla = dot_product(grad(:,1),zcglwk(:,jm))
      grad(:,1) = grad(:,1) - dla*zcglwk(:,jm)
    enddo

    ! calculate zbeta b_j+1 = || v_j+1 ||
    zbeta(iter+1) = sqrt(dot_product(grad(:,1),grad(:,1)))
    zcglwk(:,iter+1) = grad(:,1)/zbeta(iter+1)

    zqg0(iter+1,1) = dot_product(zcglwk(:,iter+1),zgrad0(:,1))

    ! calculate reduction in gradient norm
    ! ------------------------------------

    ! calculate tridiagonal matrix, T
    if ( allocated(zmat) ) deallocate(zmat)
    if ( allocated(izmat) ) deallocate(izmat)
    allocate( zmat(iter,iter) )
    allocate( izmat(iter,iter) )
    zmat(:,:) = 0.
    do i = 1, iter
      zmat(i,i) = zdelta(i)
    enddo
    if(iter.gt.1) then
      do i = 1, iter-1
        zmat(i,i+1) = zbeta(i+1)
        zmat(i+1,i) = zbeta(i+1)
      enddo
    endif
    if ( config%verbose ) then
      write(logid,*) 'zmat = '
      do i = 1, iter
        write(logid,*) zmat(i,:)
      enddo
    endif

    ! calculate inverse of T
    if ( iter.gt.1) then
      ! use eigen decomposition
      lwork = 26*iter
      liwork = 10*iter
      allocate(work(lwork))
      allocate(iwork(liwork))
      allocate(isuppz(2*iter))
      if ( allocated(zv) ) deallocate(zv)
      if ( allocated(zritz) ) deallocate(zritz)
      allocate(zv(iter,iter))
      allocate(zritz(iter,1))
      allocate(ieval(iter,iter))
      call dsyevr('V','A','L',iter,zmat,iter,1.,1.,1,1,0.,m,zritz(:,1),zv,iter,isuppz,work,lwork,iwork,liwork,info)
      if(info.ne.0) then
        write(logid,*) 'ERROR congrad: eigenvalue decomposition'
        stop
      endif
      deallocate(work)
      deallocate(iwork)
      deallocate(isuppz)
      ieval(:,:) = 0.
      do i = 1, iter 
        ieval(i,i) = 1d0/zritz(i,1)
      end do
      izmat = matmul(zv,ieval)
      izmat = matmul(izmat,transpose(zv))
      deallocate(ieval)
    else
      izmat(1,1) = 1d0/zmat(1,1)
    endif

    ! T^(-1)*(-1*zqg0)
    zwork(1:iter,1) = -1d0*zqg0(1:iter,1)
    zwork(1:iter,1) = matmul(izmat,zwork(1:iter,1))
    zw(:,1) = zgrad0(:,1) + zbeta(iter+1)*zcglwk(:,iter+1)*zwork(iter,1)
    zw(:,1) = zw(:,1) - matmul(zcglwk(:,1:iter), zqg0(1:iter,1))
    reduct = sqrt(dot_product(zw(:,1),zw(:,1)))/zgnorm
    write(logid,*) 'Reduction in norm of gradient is: ',reduct

    ! check Ritz values
    ! ---------------------
    
    if (iter.gt.1) then
      if ( minval(zritz(:,1)).lt.0 ) then
        write(logid,*) 'WARNING congrad: negative Ritz value'
        exit lanczos_iter 
      endif
    endif

    if ( iter.ge.maxiter.or.reduct.lt.reqrd) exit lanczos_iter

  end do lanczos_iter

  ! calculate solution vector and gradient
  ! --------------------------------------

  ! z = z0 + V*T^(-1)*beta(1)*e
  ! where V is matrix of lanczos vectors V = (v1,v2,...vm)
  za(:,1) = zbg(:,1) + matmul(zcglwk(:,1:iter), zwork(1:iter,1))
  
  grad(:,1) = zgrad0(:,1) + zbeta(iter+1)*zcglwk(:,iter+1)*(-1d0)*zqg0(iter,1)
  grad(:,1) = grad(:,1) - matmul(zcglwk(:,1:iter), zqg0(1:iter,1))

  ! transform state vector to physical space
  xa = chi2phi(states, covar, za)
  states%px = real(xa(:,1),kind=4)

  if (iter.eq.1) go to 10

  write(logid,*) 'zritz = ',zritz

  ! calculate posterior uncertainty
  ! -------------------------------

  ! Ba = B*(J''(chi))^(-1)
  ! J''(chi)^(-1) = A^(-1) = V*T^(-1)*V^(T)

  ! reorder ritz values
  allocate( ptheta(iter) )
  ptheta(:) = 0d0
  ngood = 0
  do i = iter, 1, -1
    if ( zritz(i,1) >= 1. ) then
      ngood = ngood + 1
      ptheta(ngood) = zritz(i,1)
    endif
  enddo
  write(logid,*) 'Number of good Ritz values: ',ngood

  ! calculate eigenvectors of Hessian
  allocate( pvcglev(nvar,ngood) ) 
  pvcglev(:,:) = 0d0

  ! equivalent of wrevecs in PyVAR
  ij = 0
  do jm = iter, 1, -1
    if ( zritz(jm,1) >= 1.) then
      ij = ij+1
      pvcglev(:,ij) = 0d0
      do jk = 1, iter
        pvcglev(:,ij) = pvcglev(:,ij) + zcglwk(:,jk)*zv(jk,jm)
      enddo
      do jk = 1, ij-1
        dla = dot_product(pvcglev(:,jk),pvcglev(:,ij))
        pvcglev(:,ij) = pvcglev(:,ij) - dla*pvcglev(:,jk)
      enddo
      dla = dot_product(pvcglev(:,ij),pvcglev(:,ij))
      pvcglev(:,ij) = pvcglev(:,ij)/sqrt(dla)
    endif
  enddo

  ! save eigen values and vectors of Hessian
  write(rowfmt,'(A,I6,A)') '(',ngood,'(F16.4,1X))'
  open(100,file=trim(files%path_output)//'hessian_evecs.txt',status='replace',action='write',iostat=ierr)
  do i = 1, nvar 
    write(100,fmt=rowfmt) pvcglev(i,:) 
  enddo  
  close(100)
  open(100,file=trim(files%path_output)//'hessian_evals.txt',status='replace',action='write',iostat=ierr)
  do i = 1, ngood
    write(100,fmt='(F16.4)') ptheta(i)
  enddo
  close(100)

  ! calculate inverse of Hessian
  allocate( pevecs(ngood,nvar) )
  pevecs(:,:) = 0d0
  do jk = 1, ngood
    pevecs(jk,:) = pvcglev(:,jk)*sqrt(1d0-1d0/ptheta(jk))
  enddo

  ! xerr^2 = xerr0^2 - sum[(1-1/eval[j])(B^(1/2)*evec(j,))^2] 
  ! calculate sum[(1-1/eval[j])(B^(1/2)*evec(j,))^2]
  tmp(:) = 0d0
  do jk = 1, ngood
    do jt = 1, ntstate
      do k = 1, ndvar
        do j = 1, neig
          tmp0(j) = covar%evecs(k,j)*sqrt(covar%evals(j))
        end do
        do i = 1, ndvar
          tmp1(i) = dot_product(tmp0,covar%evecs(i,:))
        end do
        do it = 1, ntstate
          tmp((jt-1)*ndvar+k) = tmp((jt-1)*ndvar+k) + ( covar%sqcort(jt,it) * &
                                 dot_product(tmp1, pevecs(jk,(it-1)*ndvar+1:it*ndvar)) )**2
        end do
      end do
    end do
  end do
  ! calculate last part 
  do i = 1, nvar 
    var = dble(states%pxerr0(i))**2 - tmp(i) 
    if ( var >= 0. ) states%pxerr(i) = real(sqrt(var),kind=4)
  enddo

10 continue

  if ( allocated(zmat) )    deallocate(zmat)
  if ( allocated(izmat) )   deallocate(izmat)
  if ( allocated(zv) )      deallocate(zv)
  if ( allocated(zritz) )   deallocate(zritz)  
  if ( allocated(ptheta) )  deallocate(ptheta)
  if ( allocated(pvcglev) ) deallocate(pvcglev)
  if ( allocated(pevecs) )  deallocate(pevecs)

end subroutine congrad


